#------------------------------------------------------------------------------
# ABOUT NNIpy
#------------------------------------------------------------------------------

# NNIpy is a Python package for conducting natural neighbour interpolation.  It
# is unique in providing the ability to interpolate beyond the convex hull of
# the data points.

#------------------------------------------------------------------------------
# LICENSING
#------------------------------------------------------------------------------

# The MIT License (MIT)

# Copyright (c) 2016 Thomas R. Etherington

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

#------------------------------------------------------------------------------

from scipy import spatial
import numpy as np

#------------------------------------------------------------------------------

def reflectPoints(points, extent):
    """    
    Reflection of points across planes defined by extent.

    Parameters
    ----------
    points : array
        NumPy array of shape (n, 2) where n is the number of points being
        mirrored.  Column 0 are x coordinates, and column 1 are y
        coordinates.
    extent : list
        A four element list [minX, maxX, minY, maxY] describing the minimum and
        maximum coordinates of the rectangular interpolation analytical extent,
        for which the boundary edges are used for reflection.
        
    Returns
    -------
    out : array
        NumPy array of shape (n*5, 2) where n is the number of points being
        reflected.  Column 0 are x coordinates, and column 1 are y
        coordinates.
    """
    # Extract the extent values
    xMin, xMax, yMin, yMax = extent
    # Reflect to the left
    l = np.array([xMin - (points[:,0] - xMin), points[:,1]]).T
    # Reflect to the right
    r = np.array([xMax + (xMax - points[:,0]), points[:,1]]).T    
    # Reflect to the bottom
    b = np.array([points[:,0], yMin - (points[:,1] - yMin)]).T
    # Reflect up
    u = np.array([points[:,0], yMax + (yMax - points[:,1])]).T
    # Combine original points and the reflections
    return(np.concatenate((points, l, b, r, u)))
        
#------------------------------------------------------------------------------

def polygonArea(pts):
    """    
    Calculate the area of a two-dimensional polygon.

    Parameters
    ----------
    pts : array
        NumPy array of shape (n, 2) where n is the number of vertices in the
        polygon.  Column 0 are x coordinates, and column 1 are y
        coordinates.
        
    Returns
    -------
    out : float
        The area of the polygon.
    """
    import numpy as np
    lines = np.hstack([pts,np.roll(pts,-1,axis=0)])
    area = 0.5*abs(sum(x1*y2-x2*y1 for x1,y1,x2,y2 in lines))
    return(area)

#------------------------------------------------------------------------------

def voronoiAreas(vd, n):
    """    
    Calculate the areas of the first n polygons in a Voronoi diagram.

    Parameters
    ----------
    vd : Voronoi diagram
        A scipy.spatial.qhull.Voronoi object.
    n : integer
        The first number of Voronoi polygons to calculate areas for.
        
    Returns
    -------
    out : 1D array
        The areas of the first n Voronoi polygons.
    """
    areas = []
    for i in range(n):
        v = vd.vertices[vd.regions[vd.point_region[i]]]            
        a = polygonArea(v)
        areas.append(a)
    return(np.array(areas))
    
#------------------------------------------------------------------------------

def checkDataDuplicates(xy):
    """    
    Check for duplicate locations.

    Parameters
    ----------
    xy : array
        NumPy array of shape (n, 2) where n is the number of locations.  Column
        0 are x coordinates, and column 1 are y coordinates.
        
    Returns
    -------
    out : exception
        Exception is raised when data locations are duplicated.
    """
    # Get number of locations
    n = xy.shape[0]
    # Get number of unique locations
    u = len(set(tuple([tuple(row) for row in xy])))
    if n == u:
        return
    else:
        raise Exception("ERROR: There are data points at duplicate locations")
        
#------------------------------------------------------------------------------

def checkDataLocations(xy, interp, extent):
    """    
    Check for locations beyond the extent.

    Parameters
    ----------
    data : array
        NumPy array of shape (n, 2) where n is the number of data points to
        interpolate from.  Column 0 are x coordinates, column 1 are y
        coordinates, and column 2 are data values.
    interp : array
        NumPy array of shape (n, 2) where n is the number of points to
        interpolate to.  Column 0 are x coordinates, and column 1 are y
        coordinates.
    extent : list
        A four element list describing the minimum and maximum coordinates of 
        the rectangular interpolation analytical extent.  The elements are 
        [minX, maxX, minY, maxY].
        
    Returns
    -------
    out : exception
        Exception is raised when data locations are beyond the extent.
    """
    # Combine areas
    a = np.vstack((xy, interp))
    # Get min and max locations
    xMin = np.min(a[:,0])
    xMax = np.max(a[:,0])
    yMin = np.min(a[:,1])
    yMax = np.max(a[:,1])
    xmin, xmax, ymin, ymax = extent
    if xMin > xmin and xMax < xmax and yMin > ymin and yMax < ymax:
        return
    else:
        raise Exception("ERROR: There are data and/or interpolation points not within the analytical area")
        
#------------------------------------------------------------------------------

# Natural neighbour interpolation
def nnint(data, interp, extent):
    """    
    Natural neighbour interpolation from 'data' points to 'interp' points
    within the area of 'extent'.

    Parameters
    ----------
    data : array
        NumPy array of shape (n, 3) where n is the number of data points to
        interpolate from.  Column 0 are x coordinates, column 1 are y
        coordinates, and column 2 are data values.
    interp : array
        NumPy array of shape (n, 2) where n is the number of points to
        interpolate to.  Column 0 are x coordinates, and column 1 are y
        coordinates.
    extent : list
        A four element list describing the minimum and maximum coordinates of 
        the rectangular interpolation analytical extent.  The elements are 
        [minX, maxX, minY, maxY].
        
    Returns
    -------
    out : array
        1D array of interpolated values for interp locations.
    """
    # Seperate xy locations and z value data
    xy = data[:,0:2]
    z = data[:,2]

    # Check data for issues
    checkDataDuplicates(xy)
    checkDataLocations(xy, interp, extent)

    # Build data location dictionary to enable checking of interpolation
    # points at the same location as data points
    dataDic = {}
    for i in range(xy.shape[0]):
        dataDic[(xy[i,0], xy[i,1])] = z[i]

    # Create a list to hold the results, and then process each point in turn
    interpolations = []    
    for i in range(interp.shape[0]):
        # Extract location of interpolation point
        interpi = interp[i,:]

        # Use value from data location if in same place
        if (interpi[0], interpi[1]) in dataDic.keys():
            interpolations.append(dataDic[interpi[0], interpi[1]])
        
        # otherwise interpolate
        else:
            # Do a Delaunay triangulation to find the natural neighbours
            xyi = np.concatenate((xy, [interpi]))
            dt = spatial.Delaunay(xyi)
            indices, indptr = dt.vertex_neighbor_vertices
            index = xy.shape[0]
            vt = indptr[indices[index]: indices[index + 1]]
           
            # Count the total number of points
            n = len(vt)
            # Mirror the points
            mPoints = reflectPoints(xy[vt,:], extent)
            
            # Create Voroni tesselation with reference data
            vd = spatial.Voronoi(mPoints)
            # Caluclate areas
            areas1 = voronoiAreas(vd, n)
            
            # Add interpolation point to tesselation
            vd = spatial.Voronoi(np.concatenate((mPoints, [interpi])))
            # Calulate new areas
            areas2 = voronoiAreas(vd, n)
            
            # Conduct natural neighbour interpolation
            losses = areas1 - areas2
            weights = losses / np.sum(losses)
            nnint = np.sum(weights * z[vt])        
            interpolations.append(nnint)
        
    return(np.array(interpolations))

#------------------------------------------------------------------------------
